#ifndef IMU_H
#define IMU_H
#include <Wire.h>

/*
 * The IMU indexes represent the LSB of the
 * address, which is specified by hard-wiring
 * pin AC0 to either VCC or GND.
 */
#define IMU_GENERIC_ADDR 0x68   // Address before considering AD0 pin
#define IMU_STICK_ADDR 0x69
#define IMU_STICK_IDX 1
#define IMU_BELT_ADDR 0x68
#define IMU_BELT_IDX 0
#define NUM_IMU 2

#define YAW_CORRECTION_TERM 0.0664452


#define READ_TWO_BYTES() { Wire.read() << 8 | Wire.read() }

/*
 * Define the dimension indexes
 */
#define NUM_IMU_AXIS 6
#define IMU_GYRO_X 3
#define IMU_GYRO_Y 4
#define IMU_GYRO_Z 5

/*******************
 * CONFIG REGISTER *
 *******************/
#define IMU_CONFIG_ADDR 0x1A
#define IMU_DLPF_CFG 0x0

/**********************
 * SMPRT_DIV REGISTER *
 **********************/
#define IMU_SMPRT_DIV_ADDR 0x19
#define IMU_GRYO_SMPLRT  ( IMU_DLPF_CFG == 0 ? 8000 : 1000 )

/***************************
 * GYROSCOPE CONFIGURATION *
 ***************************/
#define IMU_GYRO_ADDR 0x1B
#define IMU_GYRO_FS_SEL_0 0x0
#define IMU_GYRO_FS_SEL_1 0x8
#define IMU_GYRO_FS_SEL_2 0x10
#define IMU_GYRO_FS_SEL_3 0x18
#define IMU_GYRO_ZG_ST_1 0x20
#define IMU_GYRO_YG_ST_1 0x40
#define IMU_GYRO_XG_ST_1 0x80

/**************************
 * GYROSCOPE MEASUREMENTS *
 **************************/
#define IMU_GRYO_XOUT_15_8_ADDR 0x43
#define IMU_GRYO_XOUT_7_0_ADDR 0x44
#define IMU_GYRO_YOUT_15_8_ADDR 0x45
#define IMU_GYRO_YOUT_7_0_ADDR 0x46
#define IMU_GYRO_ZOUT_15_8_ADDR 0x47
#define IMU_GYRO_ZOUT_7_0_ADDR 0x48

/*******************************
 * ACCELEROMETER CONFIGURATION *
 *******************************/
#define IMU_ACCEL_ADDR 0x1C
#define IMU_ACCEL_XA_ST_1 0x80
#define IMU_ACCEL_YA_ST_1 0x40
#define IMU_ACCEL_ZA_ST_1 0x20
#define IMU_ACCEL_AFS_SEL_0 0x0
#define IMU_ACCEL_AFS_SEL_1 0x8
#define IMU_ACCEL_AFS_SEL_2 0x10
#define IMU_ACCEL_AFS_SEL_3 0x18

/******************************
 * ACCELEROMETER MEASUREMENTS *
 ******************************/
#define IMU_ACCEL_X 0
#define IMU_ACCEL_Y 1
#define IMU_ACCEL_Z 2
#define IMU_ACCEL_XOUT_15_8 0x3B
#define IMU_ACCEL_XOUT_7_0 0x3C
#define IMU_ACCEL_YOUT_15_8 0x3D
#define IMU_ACCEL_YOUT_7_0 0x3E
#define IMU_ACCEL_ZOUT_15_8 0x3F
#define IMU_ACCEL_ZOUT_7_0 0x40

/********************************
 * POWER MANAGEMENT 1 REGISTER *
 *******************************/
#define IMU_PWR_MGMT_1_ADDR 0x6B
#define IMU_PWR_MGMT_1_CLKSEL_0 0x0
#define IMU_PWR_MGMT_1_TEMP_DIS_1 0x8

#endif /* IMU_H */
