#include <VL53L1X.h>
#include <vl53l1_register_map.h>
#include <math.h>
#include <MadgwickAHRS.h>
#include <Wire.h>
#include "IMU.h"
#include "utils.h"

#define PLOT_IMU 0
#define ENABLE_LASER 1
#define ENABLE_IMU 0
#define ENABLE_DRIVER 1
#define ENABLE_SERIAL 0

#define LASER_1_SHUT_PIN 1

#define PRINT_LINE() \
do { \
  Serial.println(__LINE__); \
} while(0)

#define LASER_OFFLINE(i) \
do { \
  Serial.print("Laser sensor "); \
  Serial.print(i); \
  Serial.println(" offline!"); \
} while(0)

typedef struct {
  int laserDistances[NUM_LASER];
  float IMUdata[NUM_IMU][NUM_IMU_AXIS];
} SensorData;

Laser lasers[NUM_LASER] = {
  { .addr = 0 },
  { .addr = 1 },
  { .addr = 2 },
  { .addr = 3 }
};
Driver drivers[NUM_HAPTIC] = {
  { .ENpin = 4 },
  { .ENpin = 6 },
  { .ENpin = 10 },
  { .ENpin = 12 }
};
SensorData inputs;
const int hapticAddr[NUM_HAPTIC] = {5,6,7,8};
const int SDApin = 4;
const int SCLpin = 5;

#define IMU_POLLING_FREQ 25
unsigned long microsPerReading, microsPrevious[NUM_HAPTIC];
unsigned long startTime;
float yaw, roll, pitch;
float yawCorrectionCounter[NUM_HAPTIC] = {0.0, 0.0};
Madgwick mg[NUM_HAPTIC];

void setup() {
  // put your setup code here, to run once:
  Wire.begin();
  #if ENABLE_SERIAL
  Serial.begin(9600);
  #endif
  
  #if ENABLE_LASER
  pinMode(LASER_1_SHUT_PIN, OUTPUT);
  pinMode(2, OUTPUT);
  pinMode(7, OUTPUT);
  pinMode(8, OUTPUT);
  digitalWrite(LASER_1_SHUT_PIN,LOW);
  digitalWrite(2,LOW);
  digitalWrite(7,LOW);
  digitalWrite(8,LOW);

  
  
  // Check lasers initialized
  for( int i = 0; i < NUM_LASER; i++){
    switch(i) {
      case 0:
        digitalWrite(LASER_1_SHUT_PIN, HIGH);
        break;
      case 1:
        pinMode(2, INPUT);
        break;
      case 2:
        pinMode(7, INPUT);
        break;
      case 3:
        pinMode(8, INPUT);
        break;
    }
    lasers[i].dev.setTimeout(500);
    if ( lasers[i].dev.init( true ) == false )
      LASER_OFFLINE(i);
    else{
      Serial.print("LASER SENSOR "); Serial.print(i); Serial.println("ONLINE");
    }
    lasers[i].dev.setAddress((uint8_t)(20+i));

    lasers[i].dev.setDistanceMode(VL53L1X::Long);
 
    lasers[i].dev.setMeasurementTimingBudget(50000);
   
    lasers[i].dev.startContinuous(50);
   
  }
  #endif

  #if ENABLE_DRIVER
  Adafruit_DRV2605 driver1;
  Adafruit_DRV2605 driver2;
  Adafruit_DRV2605 driver3;
  Adafruit_DRV2605 driver4;

  pinMode(Driver1EN, OUTPUT);
  pinMode(Driver2EN, OUTPUT);
  pinMode(Driver3EN, OUTPUT);
  pinMode(Driver4EN, OUTPUT);
  
  pinMode(Driver1PWM, OUTPUT);
  pinMode(Driver2PWM, OUTPUT);
  pinMode(Driver3PWM, OUTPUT);
  pinMode(Driver4PWM, OUTPUT);

  digitalWrite(Driver1EN, HIGH);
  digitalWrite(Driver2EN, LOW);
  digitalWrite(Driver3EN, LOW);
  digitalWrite(Driver4EN, LOW);
  driver1.begin();
  driver1.setMode(0x07); //run autocalibration
  driver1.setMode(0x03); //set to pwm control

  digitalWrite(Driver1EN, LOW);
  digitalWrite(Driver2EN, HIGH);
  digitalWrite(Driver3EN, LOW);
  digitalWrite(Driver4EN, LOW);
  driver2.begin();
  driver2.setMode(0x07); //run autocalibration
  driver2.setMode(0x03); //set to pwm control
  
  digitalWrite(Driver1EN, LOW);
  digitalWrite(Driver2EN, LOW);
  digitalWrite(Driver3EN, HIGH);
  digitalWrite(Driver4EN, LOW);
  driver3.begin();
  driver3.setMode(0x07); //run autocalibration
  driver3.setMode(0x03); //set to pwm control
  
  
  digitalWrite(Driver1EN, LOW);
  digitalWrite(Driver2EN, LOW);
  digitalWrite(Driver3EN, LOW);
  digitalWrite(Driver4EN, HIGH);
  driver4.begin();
  driver4.setMode(0x07); //run autocalibration
  driver4.setMode(0x03); //set to pwm control

  digitalWrite(Driver1EN, LOW);
  digitalWrite(Driver2EN, LOW);
  digitalWrite(Driver3EN, LOW);
  digitalWrite(Driver4EN, LOW);
  #endif
  
  #if ENABLE_IMU
  // Configure IMU settings
  for( int imu = 0; imu < NUM_HAPTIC; imu++ ){
    unsigned char imu_addr = IMU_GENERIC_ADDR | imu;
    mg[imu].begin(IMU_POLLING_FREQ);
    Wire.beginTransmission(imu_addr);
    Wire.write(IMU_GYRO_ADDR);
    Wire.write(IMU_GYRO_FS_SEL_0);
    Wire.endTransmission(true);

    digitalWrite(0, LOW);
    
    Wire.beginTransmission(imu_addr);
    Wire.write(IMU_ACCEL_ADDR);
    Wire.write(IMU_ACCEL_AFS_SEL_0); 
    Wire.endTransmission(true);
    
    
    
    Wire.beginTransmission(imu_addr);
    Wire.write(IMU_PWR_MGMT_1_ADDR);
    // Wire.write(IMU_PWR_MGMT_1_TEMP_DIS_1);
    Wire.write(0);
    Wire.endTransmission(true);

    
    
    /*
     * In order for Madgwick algorithm to work, we need to set
     * the output rate of gyro to equal IMU_POLLING_FREQ. We do this
     * by writing the sample rate divider register.
     */
    Wire.beginTransmission(imu_addr);
    Wire.write(IMU_CONFIG_ADDR);
    Wire.write(IMU_DLPF_CFG);
    Wire.endTransmission(true);
    
    Wire.beginTransmission(imu_addr);
    Wire.write(IMU_SMPRT_DIV_ADDR);
    Wire.write( (IMU_GRYO_SMPLRT / IMU_POLLING_FREQ) - 1 );
    Wire.endTransmission(true);
  }
  
  #endif
  
  

  /*
   * The following pin enable is just for debugging purposesg
   */
  
  #if ENABLE_IMU
  // Initialize pacing variables to keep track of polling intervals
  microsPerReading = 1000000 / IMU_POLLING_FREQ;
  for(int i = 0; i < NUM_HAPTIC; i++) microsPrevious[i] = micros();
  startTime = micros();
  #endif
}

void loop() {

  
  
  #if ENABLE_LASER
  /* * * * * * * *
   * POLL LASERS *
   * * * * * * * */
  for( int i = 0; i < NUM_LASER; i++ ){
    inputs.laserDistances[i] = lasers[i].dev.read();

    #if 0
    Serial.print("LASER "); Serial.print(i); Serial.print(": ");
    Serial.println(inputs.laserDistances[i]);
    #endif
  }
 
  #endif

  #if ENABLE_IMU
  /* * * * * * *
   * POLL IMUs *
   * * * * * * */
  for(int imu = 0; imu < NUM_IMU; imu++){
    /*
     * Infer the IMU address by simply using the generic address
     * and ORing it with the imu index.
     */
    unsigned long microsNow;

    microsNow = micros();

    if( microsNow - microsPrevious[imu] >= microsPerReading){
      yawCorrectionCounter[imu] += YAW_CORRECTION_TERM;
      if( yawCorrectionCounter[imu] >= 360.0)
        yawCorrectionCounter[imu] = 0.0;
      
      unsigned char imu_addr = IMU_GENERIC_ADDR | imu;
      Wire.beginTransmission(imu_addr);
      Wire.write(IMU_ACCEL_XOUT_15_8);
      Wire.endTransmission();
      Wire.requestFrom(imu_addr, 6);
      while(Wire.available() != 6) delay(5);
      inputs.IMUdata[imu][IMU_ACCEL_X] = convertRawAcceleration(READ_TWO_BYTES());
      inputs.IMUdata[imu][IMU_ACCEL_Y] = convertRawAcceleration(READ_TWO_BYTES());
      inputs.IMUdata[imu][IMU_ACCEL_Z] = convertRawAcceleration(READ_TWO_BYTES());
      
      Wire.beginTransmission(imu_addr);
      Wire.write(IMU_GRYO_XOUT_15_8_ADDR);
      Wire.endTransmission(false);
      Wire.requestFrom(imu_addr, 6,true);
      while(Wire.available() != 6) delay(5);
      inputs.IMUdata[imu][IMU_GYRO_X] = convertRawGyro(READ_TWO_BYTES());
      inputs.IMUdata[imu][IMU_GYRO_Y] = convertRawGyro(READ_TWO_BYTES());
      inputs.IMUdata[imu][IMU_GYRO_Z] = convertRawGyro(READ_TWO_BYTES());
      
  
      mg[imu].updateIMU(inputs.IMUdata[imu][IMU_GYRO_X], 
        inputs.IMUdata[imu][IMU_GYRO_Y], 
        inputs.IMUdata[imu][IMU_GYRO_Z], 
        inputs.IMUdata[imu][IMU_ACCEL_X], 
        inputs.IMUdata[imu][IMU_ACCEL_Y], 
        inputs.IMUdata[imu][IMU_ACCEL_Z]);
  
      #if PLOT_IMU
      //Serial.print(mg[imu].getRoll()); Serial.print(",");
      float tmpYaw = mg[imu].getYaw() + yawCorrectionCounter[imu];
      if( tmpYaw >= 360.0 ) tmpYaw -= 360.0;
      //Serial.print( tmpYaw ); Serial.print(",");
      //Serial.print(mg[imu].getPitch()); Serial.print("\n");
      #endif

      #if 0
      if( mg.getYaw() <= 178.0 ){
        //Serial.println( micros() - startTime );
      }
      #endif
  
      microsPrevious[imu] += microsPerReading;

    }
  }

  float tmpYaw0 = mg[0].getYaw() + yawCorrectionCounter[0];
  if( tmpYaw0 >= 360.0 ) tmpYaw0 -= 360.0;
  float tmpYaw1 = mg[1].getYaw() + yawCorrectionCounter[1];
  if( tmpYaw1 >= 360.0 ) tmpYaw1 -= 360.0;

 float intensity[NUM_LASER] = {0,0,0,0};
 int minD = 2500;
 for(auto i: inputs.laserDistances){
  if(minD>i)
    minD=i;
 }


 
  
  #endif // ENABLE_IMU

  #if ENABLE_DRIVER
  /*
   * Haptic intensity is given as a percentage.
   */
  float intensity[NUM_HAPTIC];
  
  #if ENABLE_IMU
  motor_intensity(inputs.laserDistances, tmpYaw1 - tmpYaw0, intensity, minD);
  #else
  for( int i = 0; i < NUM_HAPTIC; i++ ){

    int calculation =  (int) (-0.128 * (float) inputs.laserDistances[i]) + 256;
    intensity[i] = max( min( calculation, 255 ), 0);
    Serial.print(intensity[i]); Serial.print(",");
    // Serial.print(inputs.laserDistances[i]); Serial.print(" "); Serial.print(calculation); Serial.print(" "); Serial.println(intensity[i]);
  }
  Serial.println("");
  #endif
  
  #if 0
  for(int i = 0; i < 4; i++){
    //Serial.print("MOTOR "); Serial.print(i); Serial.print(": "); Serial.println(intensity[i]);
    //Serial.print(tmpYaw1-tmpYaw0); Serial.print(",");
    //Serial.print(minD); Serial.print(",");

    Serial.print(intensity[0]); Serial.print(",");
    Serial.print(intensity[1]); Serial.print(",");
    Serial.print(intensity[2]); Serial.print(",");
    Serial.println(intensity[3]); 
  }
  #endif
  
  //MOTOR 1
  if( intensity[0] > 5 ){
    analogWrite(Driver1PWM, intensity[0]);
    digitalWrite(Driver1EN, HIGH);
  }
  else{
    analogWrite(Driver1PWM, 0);
    digitalWrite(Driver1EN, LOW);
  }
  //MOTOR 2
  if( intensity[1] > 5 ){
    analogWrite(Driver2PWM, intensity[1]);
    digitalWrite(Driver2EN, HIGH);
  }
  else{
    analogWrite(Driver2PWM, 0);
    digitalWrite(Driver2EN, LOW);
  }
  //MOTOR 3
  if( intensity[2] > 5 ){
    analogWrite(Driver3PWM, intensity[2]);
    digitalWrite(Driver3EN, HIGH);
  }
  else{
    analogWrite(Driver3PWM, 0);
    digitalWrite(Driver3EN, LOW);
  }
  //MOTOR 4
  if( intensity[3] > 5 ){
    analogWrite(Driver4PWM, intensity[3]);
    digitalWrite(Driver4EN, HIGH);
  }
  else{
    analogWrite(Driver4PWM, 0);
    digitalWrite(Driver4EN, LOW);
  }

  #endif
}
