  #include <math.h>
#include "utils.h"
#include <Wire.h>
#include <MadgwickAHRS.h>


const float laserOffsets[NUM_LASER] = {
    -37.5, -12.5, 12.5, 37.5
};

const int hapticOffsets[NUM_HAPTIC] = {
    -50, -25, 25, 50
};

void motor_intensity(int ld[], int yaw_rel, float *intensity, int minD){

  int md = min(minD, 2500); 
  if(yaw_rel > -50 && yaw_rel < -25) {
    intensity[0] = (-.102*md + 255);
    intensity[1] = 0;
    intensity[2] = 0;
    intensity[3] = 0;
  }
  else if(yaw_rel > -25 && yaw_rel < 0) {
    intensity[0] = 0;
    intensity[1] = (-.102*md + 255);
    intensity[2] = 0;
    intensity[3] = 0;
  }
  else if(yaw_rel > 0 && yaw_rel < 25) {
    intensity[0] = 0;
    intensity[1] = 0;
    intensity[2] = (-.102*md + 255);
    intensity[3] = 0;
  }
  else if(yaw_rel > 25 && yaw_rel < 50) {
    intensity[0] = 0;
    intensity[1] = 0;
    intensity[2] = 0;
    intensity[3] = (-.102*md + 255);
  }

}

int setDriverRealtimeValue( Driver drivers[NUM_HAPTIC], int driverIdx, unsigned char value ){
    for(int i = 0; i < NUM_HAPTIC; i++ ){
        if ( i != driverIdx) {
            digitalWrite(drivers[i].ENpin, LOW);
        }
    }
    drivers[driverIdx].drv.setRealtimeValue(value);
    for(int i = 0; i < NUM_HAPTIC; i++ ){
        digitalWrite(drivers[i].ENpin, HIGH);
    }
}

void getReadings(Madgwick mg, float *roll, float *pitch, float *yaw, float ax, float ay, float az, float gx, float gy, float gz)
{
    mg.updateIMU(gx, gy, gz, ax, ay, az);
    *roll = mg.getRoll();
    *pitch = mg.getPitch();
    *yaw = mg.getYaw();
}

float convertRawAcceleration(int aRaw) {
  // since we are using 2G range
  // -2g maps to a raw value of -32768
  // +2g maps to a raw value of 32767
  
  float a = (aRaw * 2.0) / 32768.0;
  return a;
}

float convertRawGyro(int gRaw) {
  // since we are using 250 degrees/seconds range
  // -250 maps to a raw value of -32768
  // +250 maps to a raw value of 32767
  
  float g = (gRaw * 250.0) / 32768.0;
  return g;
}
